/*
++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

  Author: Linus Nielsen <Linus.Nielsen@sth.frontec.se>

  Description:
  
  This file contains functions implementing a double linked list. 
  The functions uses the types LIST and LISTNODE.
  There is a trick with the three nodes in LIST. By placing the 
  tail_pred field in middle, makes it possible to avoid special code
  to handle nodes in the beginnig or end of the list.
  
  The 'head' and 'tail' field in LIST are never NULL, even if the
  list is empty.
  
  The 'succ' and 'pred' field in a LIST_NODE that is in the list,
  are never NULL. The 'pred' field for the first LIST_NODE points
  to the 'head' field in LIST. The 'succ' node for the last LIST_NODE
  points the the tail_pred field in LIST.
  
++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
*/
#include "types.h"
#include "lists.h"

/****************************************************************************
** FUNCTION:	InitList
**
** DESCRIPTION:	Initiate a LIST structure. 
**              If max_num_nodes is set to NO_SIZE_CHECK there 
**		will be no check of the length of the list.
**
** RETURN:	Nothing
******************************************************************************/
void InitList(LIST *list,		/* The list to initiate */
	      unsigned long max_num_nodes)	/* Maximum number of nodes */
{
  list->first = (LIST_NODE *)&list->last_prev;
  list->last_prev = NULL;
  list->last = (LIST_NODE *)&list->first;
  list->num_nodes = 0;
  list->max_num_nodes = max_num_nodes;
  
} /* End InitList */

/****************************************************************************
** FUNCTION:	InsertNodeBefore
**
** DESCRIPTION:	Insert a LIST_NODE in a list before another node in a LIST.
**
** RETURN:	Nothing
******************************************************************************/
void InsertNodeBefore(LIST_NODE *ref,	/* The reference node */
		      LIST_NODE *ln)	/* The node to insert */
{
  ln->next = ref;		/* ref is after us */
  ln->prev = ref->prev;		/* ref's prev is before us */
  ref->prev = ln;		/* we are before ref */
  ln->prev->next = ln;		/* we are after ref's prev */
}

/****************************************************************************
** FUNCTION:	InsertNodeAfter
**
** DESCRIPTION:	Insert a LIST_NODE in a list after another node in a LIST.
**
** RETURN:	Nothing
*****************************************************************************/
void InsertNodeAfter(LIST_NODE *ref,	/* The reference node */
		     LIST_NODE *ln)	/* The node to insert */
{
  ln->prev = ref;		/* ref is before us */
  ln->next = ref->next;		/* ref's next is after us */
  ref->next = ln;		/* we are after ref */
  ln->next->prev = ln;		/* we are before ref's next */
}

/****************************************************************************
** FUNCTION:	RemNode
**
** DESCRIPTION:	Extract a LIST_NODE from a list.
**
** RETURN:	The same LIST_NODE pointer that was passed as a parameter.
*****************************************************************************/
LIST_NODE *RemNode(LIST_NODE *ln)	/* The node to extract */
{
  ln->prev->next = ln->next;      /* Our prev's next points to our next */
  ln->next->prev = ln->prev;      /* Our next's prev points to our prev */
  return(ln);
}

/******************************************************************************
** FUNCTION:	AddFirst
**
** DESCRIPTION:	Add a LIST_NODE at the beginning of a LIST.
**
** RETURN:	TRUE if OK
**		FALSE if list was full
******************************************************************************/
BOOL AddFirst(LIST *list,		/* The list to add to */
		 LIST_NODE *ln)		/* The node to add */
{
  if (NO_SIZE_CHECK != list->max_num_nodes)
  {
    if(list->num_nodes >= list->max_num_nodes)	/* List full? */
    {
      return(FALSE);
    }
  }
  InsertNodeAfter((LIST_NODE *)list, ln);
  list->num_nodes++;		/* Increment node counter */
  return(TRUE);
}

/******************************************************************************
** FUNCTION:	RemFirst
**
** DESCRIPTION:	Extract a LIST_NODE from the beginning of a LIST.
**
** RETURN:	The extracted LIST_NODE or NULL if the list is empty
******************************************************************************/
LIST_NODE *RemFirst(LIST *list)	/* The list to extract from */
{
  LIST_NODE *ln;
  
  if(ListEmpty(list))		/* Return NULL if the list is empty */
  {
    return(NULL);
  }
  
  ln = RemNode((LIST_NODE *)list->first);  /* Get first node */
  
  list->num_nodes--;			/* Decrement node counter */
  
  return(ln);
}

/******************************************************************************
** FUNCTION:	AddLast
**
** DESCRIPTION:	Add a LIST_NODE at the end of a LIST.
**
** RETURN:	NULL if OK
**		FALSE if list was full
******************************************************************************/
BOOL AddLast(LIST *list,		/* The list to add to */
		LIST_NODE *ln)		/* The node to add */
{
  if (NO_SIZE_CHECK != list->max_num_nodes)
  {
    if(list->num_nodes >= list->max_num_nodes)	/* List full? */
    {
      return(FALSE);      
    }
  }
  InsertNodeBefore((LIST_NODE *)&list->last_prev, ln);
  list->num_nodes++;			/* Increment node counter */
  return(TRUE);
}

/******************************************************************************
** FUNCTION:	RemLast
**
** DESCRIPTION:	Extract a LIST_NODE from the end of a LIST.
**
** RETURN:	The extracted LIST_NODE or NULL if the list is empty
******************************************************************************/
LIST_NODE *RemLast(LIST *list)	/* The list to extract from */
{
  LIST_NODE *ln;
  
  if(ListEmpty(list))		/* Return NULL if the list is empty */
  {
    return(NULL);
  }
  
  ln = RemNode((LIST_NODE *)list->last);
  
  list->num_nodes--;	/* Decrement node counter */
  
  return(ln);		/* Is NULL if the list is empty */
}

/******************************************************************************
** FUNCTION:	LastInList
**
** DESCRIPTION:	Check if a LIST_NODE is last in the list
**
** RETURN:	TRUE if last in list
******************************************************************************/
BOOL LastInList(LIST_NODE *ln)	/* The node to check */
{
  return ln->next->next == NULL;
}

/*****************************************************************************
** FUNCTION:	FirstInList
**
** DESCRIPTION:	Check if a LIST_NODE is first in the list
**
** RETURN:	TRUE if first in list
******************************************************************************/
BOOL FirstInList(LIST_NODE *ln)	/* The node to check */
{
  return ln->prev->prev == NULL;
}

/******************************************************************************
** FUNCTION:	ListEmpty
**
** DESCRIPTION:	Check if a LIST is empty
**
** RETURN:	TRUE if list is empty
******************************************************************************/
BOOL ListEmpty(LIST *list)		/* The list to check */
{
  return list->first == (LIST_NODE *)&list->last_prev;
}

/******************************************************************************
** FUNCTION:	GetFirst
**
** DESCRIPTION:	Return a LIST_NODE from the beginning of a LIST.
**
** RETURN:	The first LIST_NODE or NULL if the list is empty
******************************************************************************/
LIST_NODE *GetFirst(LIST *lh)	/* The list to read from */
{
  if(ListEmpty(lh))	/* Return NULL if the list is empty */
  {
    return(NULL);
  }
  
  return lh->first;	/* Get first node */
}

/******************************************************************************
** FUNCTION:	GetLast
**
** DESCRIPTION:	Return a LIST_NODE from the end of a LIST.
**
** RETURN:	The last LIST_NODE or NULL if the list is empty
******************************************************************************/
LIST_NODE *GetLast(LIST *lh)	/* The list to read from */
{
  if(ListEmpty(lh))		/* Return NULL if the list is empty */
  {
    return(NULL);
  }
  
  return lh->last;
}

/******************************************************************************
** FUNCTION:	GetNext
**
** DESCRIPTION:	Return the LIST_NODE following the specified one.
**
** RETURN:	Next LIST_NODE or NULL if the list ends here
*******************************************************************************/
LIST_NODE *GetNext(LIST_NODE *ln) /* The list node to get next from */
{
  if(LastInList(ln))		/* Return NULL if this is the end of list */
  {
    return(NULL);
  }
  
  return ln->next;
}

/******************************************************************************
** FUNCTION:	GetPrev
**
** DESCRIPTION:	Return the LIST_NODE preceding the specified one.
**
** RETURN:	Previous LIST_NODE or NULL if the list ends here
*******************************************************************************/
LIST_NODE *GetPrev(LIST_NODE *ln) /* The list node to get previous from */
{
  if(FirstInList(ln))		/* Return NULL if this is the start of list */
  {
    return(NULL);
  }
  
  return ln->prev;
}
