#ifndef LISTS_INCLUDED
#define LISTS_INCLUDED

/*
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++  
*   		DESCRIPTION					    *
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++  
	   
  Author: Linus Nielsen <Linus.Nielsen@sth.frontec.se>
  
  Functions for handling of a double linked list (code in lists.c):
   
  InitList		- Initiate a LIST structure
  InsertNodeBefore	- Insert a node in a list before another node
  InsertNodeAfter	- Insert a node in a list after another node
  RemNode		- Extract a node from a list
  RemFirst		- Extract a node from the beginning of a list
  RemLast		- Extract a node from the end of a list
  AddFirst		- Add a node at the beginning of a list
  AddLast		- Add a node at the end of a list
  LastInList		- Check if a node is last in the list
  FirstInList		- Check if a node is first in the list
  ListEmpty		- Check if a list is empty
  GetFirst		- Returns the first node in the list
  GetLast		- Returns the last node in the list
  GetNext		- Returns the next node
  GetPrev		- Returns the previous node
   
 General about the list functions:
 
 This package contains two structs and a set of functions implementing a
 double linked list. When using a list with these, add nodes with
 AddFirst or AddLast and extract nodes with RemFirst or RemLast.
 
 A FIFO queue can be implemented using AddLast to insert
 nodes at the end and RemFirst to extract nodes at the beginning
 of the list.

 When using the list-functions you need one instance of the type LIST
 (contains the start pointer etc), You also need to call InitList once.
 When calling InitList you decide if the list shall have a max size,
 and if so how big it will be.
 
 NOTE! when using InsertNodeBefore, InsertNodeAfter and
 RemNode the num_nodes field in the LIST struct will not be updated.
 
 To use the list-functions you define a struct with LIST_NODE as one member.
 When a LIST_NODE is extracted from the list the address of the LIST_NODE 
 is returned. If the LIST_NODE field is not the first field in your struct,
 you must subtract the returned address with the size of the field that is 
 before LIST_NODE in your struct to get the start address of your struct.
    
 For example: Consider you want to use the list-functions to implement a FIFO
 queue for the test_list_sig signal, using the OSE operating system.
 
 You may declare your signal struct something like:
 
    typedef struct
    {
       SIGSELECT sig_no;
       LIST_NODE list_node;
       .
       .
    } TEST_LIST_STRUCT;
    
 In your code you may have:
 
    LIST test_list;
    TEST_LIST_STRUCT *list_sig;
    LIST_NODE *ln;
        
    InitList(&test_list, MAX_NODES_IN_LIST);
    
    for(;;)
    {
        sig = receive(all);
	switch(sig->sig_no)
	{
	   case TEST_ADD:
	      if (AddFirst(&test_list, &sig->list_node))
	      {
	          * The list was full *
		  * Handle the returned sig *
	      }
	      
	   case TEST_REM:
		if (ln = RemLast(&test_list))
		{
	          list_sig = (TEST_LIST_STRUCT *)((char)ln - sizeof(SIGSELECT));
		  * Continue ... *
		}
		else
		{
		   * handle empty list *
		   
*/

/*--------------------------------------------------------------------
   List handling definitions
----------------------------------------------------------------------*/

/* Used as input parameter to InitList */
#define NO_SIZE_CHECK 0




/***********************************************************************
** Type definitions
**
** The LIST_NODE, and LIST types are used by the LIST class in common drv
************************************************************************/


typedef struct list_node
{
   				/*** NOTE! The next and prev fields 
				     are never NULL within a LIST.
        			     Use FirstInList and 
				     LastInList to check if a node 
				     is first or last in the list     ***/
  struct list_node *next;	     /* Successor node 			*/
  struct list_node *prev;	     /* Predessor node 			*/
} LIST_NODE;

typedef struct
{
  /* The three list node fields are 
     used for internal represenation 
     of the list.				*/
  LIST_NODE *first;		/* First node in list 			*/
  LIST_NODE *last_prev;		/* Always NULL 				*/
  LIST_NODE *last;		/* Last node in list 			*/
  ULONG num_nodes;		/* Number of nodes in list 		*/
  ULONG max_num_nodes;		/* Max number of nodes in list 		*/
} LIST;


/*=======================================================================*
********              PROTOTYPES                  		      ****
========================================================================*/

/******************************************************************************
**
** FUNCTION:	InitList
**
** DESCRIPTION:	Initiate a LIST structure.
**
** INPUT:	The LIST to be initiated
**	
**		Max nr of elements in the list, or NO_SIZE_CHECK if 
**		list may be of any size
*******************************************************************************
*/
void InitList(LIST *list,		/* The list to initiate */
	      ULONG max_num_nodes);	/* Maximum number of nodes */

/******************************************************************************
**
** FUNCTION:	InsertNodeBefore
**
** DESCRIPTION:	Insert a LIST_NODE before another LIST_NODE in a LIST.
**
** INPUT:	The reference LIST_NODE
**
**		The LIST_NODE to be inserted
**
** OBSERVE:     When using InsertNodeBefore, InsertNodeAfter
**              and RemNode the nume_nodes field in the LIST struct
**              will not be updated.
******************************************************************************
*/
void InsertNodeBefore(LIST_NODE *ref,	/* The reference node */
		      LIST_NODE *ln);	/* The node to insert */

/*****************************************************************************
**
** FUNCTION:	InsertNodeAfter
**
** DESCRIPTION:	Insert a LIST_NODE after another LIST_NODE in a LIST.
**
** INPUT:	The reference LIST_NODE
**
**		The LIST_NODE to be inserted
**
** OBSERVE:     When using InsertNodeBefore, InsertNodeAfter
**              and RemNode the num_nodes field in the LIST struct
**              will not be updated.
*******************************************************************************
*/
void InsertNodeAfter(LIST_NODE *ref,	/* The reference node */
		     LIST_NODE *ln);	/* The node to insert */

/******************************************************************************
**
** FUNCTION:	RemNode
**
** DESCRIPTION:	Extract a LIST_NODE from a LIST.
**
** INPUT:	The LIST_NODE to be removed.
**
** RETURN:	The same LIST_NODE pointer that was passed as a parameter
**
** OBSERVE:     When using InsertNodeBefore, InsertNodeAfter
**              and RemNode the num_nodes field in the LIST struct
**              will not be updated.
*******************************************************************************
*/
LIST_NODE *RemNode(LIST_NODE *ln);	/* The node to extract */

/******************************************************************************
**
** FUNCTION:	AddFirst
**
** DESCRIPTION:	Add a LIST_NODE at the beginning of a LIST.
**
** INPUT:	The LIST
**
**		The LIST_NODE 
**
** RETURN:	TRUE if OK
**		FALSE if list is full
*******************************************************************************
*/
BOOL AddFirst(LIST *list,	/* The list to add to */
	      LIST_NODE *ln);	/* The node to add */

/******************************************************************************
**
** FUNCTION:	RemFirst
**
** DESCRIPTION:	Extract a LIST_NODE from the beginning of a LIST.
**
** INPUT:	The LIST from where a node is to be removed.
**
** RETURN:	The extracted LIST_NODE or NULL if the list is empty
*******************************************************************************
*/
LIST_NODE *RemFirst(LIST *list);	/* The list to extract from */

/******************************************************************************
**
** FUNCTION:	AddLast
**
** DESCRIPTION:	Add a LIST_NODE at the end of a LIST.
**
** INPUT:	The LIST
**
**		The LIST_NODE 
**
** RETURN:	TRUE if OK
**		FALSE if list is full
*******************************************************************************
*/
BOOL AddLast(LIST *list,	/* The list to add to */
	     LIST_NODE *ln);	/* The node to add  */
/******************************************************************************
**
** FUNCTION:	RemLast
**
** DESCRIPTION:	Extract a LIST_NODE from the end of a LIST.
**
** INPUT:	The LIST from where a node is to be removed.
**
** RETURN:	The extracted LIST_NODE or NULL if the list is empty
*******************************************************************************
*/
LIST_NODE *RemLast(LIST *list);		/* The list to extract from */

/******************************************************************************
**
** FUNCTION:	LastInList
**
** DESCRIPTION: Check if a LIST_NODE is at the end of the list
**
** INPUT:	The LIST_NODE to check
**
** RETURN:	TRUE if LIST_NODE is last in the list, else FALSE.
*******************************************************************************
*/
BOOL LastInList(LIST_NODE *ln);		/* The node to check */

/******************************************************************************
**
** FUNCTION:	FirstInList
**
** DESCRIPTION: Check if a LIST_NODE is first in the list
**
** INPUT:	The LIST_NODE to check
**
** RETURN:	TRUE if LIST_NODE is first in the list, else FALSE.
*******************************************************************************
*/
BOOL FirstInList(LIST_NODE *ln);	/* The node to check */

/******************************************************************************
**
** FUNCTION:	ListEmpty
**
** DESCRIPTION: Check if a LIST is empty
**
** INPUT:	The LIST to check
**
** RETURN:	TRUE if LIST is empty, else FALSE.
**
*******************************************************************************
*/ 
BOOL ListEmpty(LIST *list);	/* The list to check */

/******************************************************************************
** FUNCTION:	GetFirst
**
** DESCRIPTION:	Return a LIST_NODE from the beginning of a LIST.
**
** RETURN:	The first LIST_NODE or NULL if the list is empty
******************************************************************************/
LIST_NODE *GetFirst(LIST *lh);	/* The list to read from */

/******************************************************************************
** FUNCTION:	GetLast
**
** DESCRIPTION:	Return a LIST_NODE from the end of a LIST.
**
** RETURN:	The last LIST_NODE or NULL if the list is empty
******************************************************************************/
LIST_NODE *GetLast(LIST *lh);	/* The list to read from */

/******************************************************************************
** FUNCTION:	GetNext
**
** DESCRIPTION:	Return the LIST_NODE following the specified one.
**
** RETURN:	Next LIST_NODE or NULL if the list ends here
*******************************************************************************/
LIST_NODE *GetNext(LIST_NODE *ln); /* The list node to get next from */

/******************************************************************************
** FUNCTION:	GetPrev
**
** DESCRIPTION:	Return the LIST_NODE preceding the specified one.
**
** RETURN:	Previous LIST_NODE or NULL if the list ends here
*******************************************************************************/
LIST_NODE *GetPrev(LIST_NODE *ln); /* The list node to get previous from */

#endif /* LISTS_INCLUDED */
